import numpy as np
import deepxde as dde
import scipy.io
import os
import tensorflow as tf
import matplotlib.pyplot as plt
from scipy.integrate import quad
from deepxde.callbacks import EarlyStopping

# Define global parameters
nu = 0.01

def iniFunc(x):
    mu = x[:,1:2]
    return np.sin(np.pi*mu*x[:,0:1])


def boundary_l(x, on_boundary):
    return on_boundary and np.isclose(x[0], 0)


def boundary_r(x, on_boundary):
    return on_boundary and np.isclose(x[0], 1)


def boundary_d(x, on_boundary):
    return on_boundary and np.isclose(x[1], 1)


def boundary_u(x, on_boundary):
    return on_boundary and np.isclose(x[1], 3)


def dirBC(X):
    def a0(x, mu):
        return np.exp(-1*(2 * np.pi * mu * nu)**(-1) * (1 - np.cos(mu * np.pi * x)))

    def an(x, mu, n):
        return np.exp(-1*(2 * np.pi * mu * nu)**(-1) * (1 - np.cos(mu * np.pi * x))) * np.cos(mu * n * np.pi * x)

    x_range = X[:,0:1]
    mu_range = X[:,1:2]
    t_range = X[:,2:]

    u = np.empty(x_range.shape)

    for i in range(x_range.shape[0]):
        x = x_range[i]
        mu = mu_range[i]
        t = t_range[i]
        a0val = quad(a0, 0, 1, args=(mu,))[0]
        nTerms = 100
        num = 0
        denom = a0val

        for n in range(1,nTerms):
            anval = 2*quad(an, 0, 1, args=(mu,n))[0]

            num = num + anval*np.exp(-n**2*np.pi**2*mu**2*nu*t)*n*np.sin(mu*n*np.pi*x)
            denom = denom + anval*np.exp(-n**2*np.pi**2*mu**2*nu*t)*np.cos(mu*n*np.pi*x)
        u[i] = 2*np.pi*mu*nu*num/denom
    return u


def hardBC(X, y):
    # This is a test on how to strongly enforce Dirichlet BCs
    x = X[:, 0:1]
    x_new = 0

    return x_new


def load_data(mu):
    pathData = '/Users/guillembarroso/Google Drive/GB-MG-AH/Python/deepxde_mod/examples/dataset/data_mu{}_sin_ok.mat'.format(float(mu), '.1f')
    data = scipy.io.loadmat(pathData)
    data = data['data']
    X = data[:,0:2]
    y = data[:, 2:]
    return X, y


def main():
    def pde(x, y):
        print('PDE function...')
        dy_x = dde.grad.jacobian(y, x, i=0, j=0)
        dy_t = dde.grad.jacobian(y, x, i=0, j=1)
        dy_xx = dde.grad.hessian(y, x, i=0, j=0)
        return dy_t + y * dy_x - 0.01 * dy_xx

    geom = dde.geometry.Rectangle([0,1], [1,3])
    timedomain = dde.geometry.TimeDomain(0,1)
    geomtime = dde.geometry.GeometryXTime(geom, timedomain)

    # Boundary conditions
    #bc = dde.DirichletBC(geomtime, lambda x: 0, lambda _, on_boundary: on_boundary) # same BC in all boundaries
    bc_l = dde.DirichletBC(geomtime, lambda x: 0, boundary_l)
    bc_r = dde.DirichletBC(geomtime, dirBC, boundary_r)
    bc_d = dde.DirichletBC(geomtime, dirBC, boundary_d)
    bc_u = dde.DirichletBC(geomtime, dirBC, boundary_u)

    # Parametric initial condition
    ic = dde.IC(geomtime, iniFunc, lambda _, on_initial: on_initial)

    # Select data for training
    data = dde.data.TimePDE(geomtime, pde, [bc_l, bc_r, bc_d, bc_u, ic], num_domain=2500, num_boundary=80, num_initial=160, num_test=2500)
    # dde.data.PDE(..., anchors=X) train with matrix X of dimensions N x d, N points of dimension d.

    # Define NN architecture
    net = dde.maps.FNN([3] + [20] * 3 + [1], "tanh", "Glorot normal")
    model = dde.Model(data, net)

    # net.apply_output_transform(hardBC)

    # Train model
    training = True
    model.compile("adam", lr=1e-3, loss_weights=[1,1,1,1])
    if training == True:
        early_stopping = EarlyStopping(min_delta=1e-5, patience=5000)
        losshistory, train_state = model.train(epochs=30000, callbacks=[early_stopping],disregard_previous_best=True,
                                               model_save_path='/Users/guillembarroso/Google Drive/GB-MG-AH/Python/deepxde_mod/models/model.ckpt')
        dde.saveplot(losshistory, train_state, issave=True, isplot=True)
    else:
        model.restore("/Users/guillembarroso/Google Drive/GB-MG-AH/Python/deepxde_mod/models/model.ckpt-30000")

    # the first column is PDE loss, the second column is left BC loss, and the third column is right BC loss, 4th column initial condition
    # The train loss includes all the loss terms, and the test loss only includes the PDE loss

    X, y_true = load_data(mu)
    y_pred = model.predict(X)
    f = model.predict(X, operator=pde)

    tstep = 0
    At = 1e-2
    npoints = 201

    xplot = X[npoints*tstep:npoints + npoints*tstep,0]
    ypred = y_pred[npoints*tstep:npoints + npoints*tstep]
    ytrue = y_true[npoints*tstep:npoints + npoints*tstep]
    fplot = f[npoints*tstep:npoints + npoints*tstep]

    rMean = np.mean(np.absolute(fplot))
    l2rel = dde.metrics.l2_relative_error(ytrue, ypred)
    print("Mean residual:", rMean)
    print("L2 relative error:", l2rel)

    plt.figure()
    plt.plot(xplot,ypred,label='$u_{NN}(x,t)$')
    plt.plot(xplot,ytrue,label='$u(x,t)$')
    plt.plot(xplot, fplot,label='residual')
    plt.title('$t = {}$ s, $mean_r$ = {:.4f}, $L2_u$ = {:.4f}'.format(At*tstep, rMean, l2rel))
    plt.xlabel("x")
    plt.legend()
    plt.show()

if __name__ == "__main__":
    main()
